/*
 * plc4
 *  yew plc (fa-m3) library
 *   k.furukawa, oct.6.1997, oct.25.1997, jul.20.1999.
 *
 * k.f sep.19.2002.  correct protocol, retry on timout as well (v3.0)
 * k.f jul.3.2003.   add query protocol (v3.1)
 * k.f aug.10.2003.  merge WRD/WWR protocol as well as MCM (v3.2)
 * k.f aug.17.2006.  add tcp option (v3.3)
 */
/*
   this program is a test program.
 */
static char what_plc4[] =
"@(#)plc4 v3.3.0 yew plc, k.furukawa, oct.1997-aug.2006";

#include <stdio.h>
#include <stdlib.h>			/* strtol */
#include <limits.h>			/* INT_MAX */
#include <ctype.h>			/* isprint */
#include "comsub.h"

#define MAX_MSG 2048
#define MAX_QUERY 48
#define YEW_PL_PORT 0x3001		/* paso-com link port */
#define RESP_TIMEOUT 1
#define RESP_TIME_MS (RESP_TIMEOUT*1000)

#ifdef DEBUG_ON
static int s2opt = S2_OALL;
static int debugon = 1;
#else
static int s2opt = S2_OERR;
static int debugon = 0;
#endif

static int sd = -1;			/* socket descriptor */
static char hostsd[64];			/* saved hostname */
static unsigned short resp[MAX_MSG/2];	/* response from plc */

/**********
 * plcclose -- close a channel to a plc
 */
int plcclose
()
{
    if ( sd != -1 )  {
	s2_close(sd,s2opt);
    }
    return 0;
}

/**********
 * plcwrite4 -- write values into a plc (MCM)
 */
int plcwrite4
(
 char *host,	/* host   - host name or ip address of yew plc */
 char *caddr,	/* caddr  - address to write ("Xxxxx") */
 int addr,	/* addr   - address to write (Dxxxx), when caddr is NULL */
 char *trans,	/* trans  - transport to use */
 int size,	/* size   - number of words to write (maximum of 247?) */
 unsigned short *data	/* data   - data to write */
 )
/* externals: sd, resp, hostsd, s2opt, debugon */
{
    static char template[] = {			/* command to plc */
	0x71,	/* 0: write words sequentially (large number)	*/
	0x01,	/* 1: CPU No.					*/
	0x00,	/* 2: sentence consisting of (4 + 10 + 2n) bytes*/
	0x00,	/* 3: 14 + n*2  ( cmnd[4] -> cmnd[-] )	*/
	0x26, 0x0a,
	0x00,		/* 6: data size */
	0x00,		/* 7: 10 + n*2 */
	0x00, 0x02,
	0x00,		/* 10: data number */
	0x00,		/* 11: n */
	0x00, 0x04,	/* 12: class of device = D (0x0004)	*/
	0, 0, 
	0x03, 0xe8,	/* 16: start at channel, ex. No.00005000 */
	0x00, 0x00	/* 18: data				*/
    };
    static unsigned short cmnd[MAX_MSG/2];	/* command for plc */
    int slen = size*2+18;			/* command length */
    int rlen;					/* reply length */

    if ((size>MAX_MSG/4)||(size<1))  goto E_SIZE;

    if ( cmnd[0] != htons(0x7101) )  {		/* first time */
	cmnd[0] = htons(0x7101);		/* command & cpu number */
	cmnd[2] = htons(0x260a);
	cmnd[4] = htons(0x0002);
	cmnd[7] = htons(0x0000);		/* data address uppper word */
    }
    if ( sd == -1 || strncmp(host,hostsd,64) )  {
	s2_close(sd,S2_OERR);
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt )) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	strncpy(hostsd,host,64);		/* save hostname */
    }
    
    cmnd[1] = htons(4+10+size*2);		/* message size */
    cmnd[3] = htons(10+size*2);			/* data size */
    cmnd[5] = htons(size);			/* data number */
     
    if (caddr != NULL)  {			/* address in "Xxxxx" format */
	char c;
	int d;
	if ( 2 != sscanf(caddr, "%c%d", &c, &d) )  
	    goto E_ADDR;			/* address format */
	cmnd[6] = htons(c-'@');			/* data class */
	cmnd[8] = htons(d);			/* data address */
    }  else  {					/* Dxxxx address */
	cmnd[6] = htons(0x0004);		/* data class */
	cmnd[8] = htons(addr);			/* data address */
    }	
	    
    {
	int i;
	for (i=0; i<size; i++)  {
	    cmnd[i+9] = htons(data[i]);
	}
    }

    if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	!= slen )  {
	s2_close(sd,S2_OERR);			/* try again */
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt)) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	    != slen )  {
	    if (rlen == S2_TIMEOUT)  {		/* time out ? */
		goto E_TIMEOUT;			/* time out error */
	    }  else  {
		goto E_NETWORK;			/* other network error */
	    }
	}
    }
    if ((rlen = s2_treceive(sd,(char *)resp,MAX_MSG,RESP_TIME_MS,s2opt))
	< 0)  {
	if (rlen == S2_TIMEOUT)
	    goto E_TIMEOUT;			/* time out error */
	else
	    goto E_NETWORK;			/* other network error */
    }
    /* s2_close(sd,s2opt); */
    
    if(ntohs(resp[0]) != 0xf100){		/* failure reponse */
	goto E_NETWORK2;			/* other network error */
    }
    return 0;

 E_SIZE:
    if(debugon)
	printf("plcwrite -- illegal data size '%d'\n", size);
    return -1;
    
 E_ADDR:
    if(debugon)
	printf("plcwrite -- illegal address specification '%s'\n", caddr);
    return -2;
    
 E_OPEN:
    if(debugon)
	printf("plcwrite -- open failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -3;
    
 E_TIMEOUT:
    if(debugon)
	printf("plcwrite -- timeout at send on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -4;

 E_NETWORK:
    if(debugon)
	printf("plcwrite -- plc network failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -5;

 E_NETWORK2:
    if(debugon)  {
	printf("plcwrite -- plc network internal failure on '%s'\n", host);
	printf(" -- terminal code is %x\n", ntohs(resp[0])&0xff);
    }
    s2_close( sd, s2opt );
    sd = -1;
    return -6;
}

/**********
 * plcread4 -- read values from a plc (MCM)
 */
int plcread4
(
 char *host,	/* host   - host name or ip address of yew plc */
 char *caddr,	/* caddr  - address to read ("Yxxxx") */
 int addr,	/* addr   - address to read (Dxxxx), when caddr is NULL */
 char *trans,	/* trans  - transport to use */
 int size,	/* size   - number of words to read (maximum of 247?) */
 unsigned short *data	/* data   - data to read */
 )
{
    static char template[] = {			/* command to plc */
	0x71,	/* 0: read words sequentially (large number)	*/
	0x01,	/* 1: CPU No.					*/
	0x00,	/* 2: sentence consisting of 18 bytes*/
	/* before aug.2002, it was said that this size should be 0x12 */
	/* but it was corrected to be 0x0e */
	/* 0x12, */ /* 3: 18 ??? ( cmnd[4] -> cmnd[-] )	*/
	0x0e,	/* 3: 14 ( cmnd[4] -> cmnd[-] )	*/
	0x26, 0x09,
	0x00,		/* 6: data size */
	0x0a,		/* 7: 10 */
	0x00, 0x02,
	0x00,		/* 10: data number */
	0x00,		/* 11: n */
	0x00, 0x04,	/* 12: kind of device = D (0x0004)	*/
	0, 0, 
	0x03, 0xe8,	/* 16: start at channel, ex. No.00005000 */
	0x00, 0x00	/* 18: data				*/
    };
    static unsigned short cmnd[MAX_MSG/2]; /* command for plc */
    int slen = 18;				/* command length */
    int rlen;					/* reply length */

    if ((size>MAX_MSG/4)||(size<1))  goto E_SIZE;

    if ( cmnd[0] != htons(0x7101) )  {		/* first time */
	cmnd[0] = htons(0x7101);		/* command & cpu number */
	cmnd[1] = htons(0x000e);		/* message size */
	cmnd[2] = htons(0x2609);
	cmnd[3] = htons(0x000a);		/* data size */
	cmnd[4] = htons(0x0002);
	cmnd[7] = htons(0x0000);		/* data address upper word */
    }
    if ( sd == -1 || strncmp(host,hostsd,64) )  {
	s2_close(sd,S2_OERR);
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt )) < 0)  {
	    goto E_OPEN;
	}
	strncpy(hostsd,host,64);		/* save hostname */
    }

    cmnd[5] = htons(size);			/* data number */

    if (caddr != NULL)  {			/* address in "Xxxxx" format */
	char c;
	int d;
	if ( 2 != sscanf(caddr, "%c%d", &c, &d) )
	    goto E_ADDR;			/* address format */
	cmnd[6] = htons(c-'@');			/* data class */
	cmnd[8] = htons(d);			/* data address */
    }  else  {
	cmnd[6] = htons(0x0004);		/* data class */
	cmnd[8] = htons(addr);			/* data address */
    }

    if (debugon >= 2) plcdumpdata(cmnd,slen,"messge to be sent");
    
    if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	!= slen)  {
	s2_close(sd,S2_OERR);			/* try again */
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt )) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	    != slen)  {
	    if (rlen == S2_TIMEOUT)  {		/* time out ? */
		goto E_TIMEOUT;			/* time out error */
	    }  else  {
		goto E_NETWORK;			/* other network error */
	    }
	}
    }
    if ((rlen = s2_treceive(sd,(char *)resp,MAX_MSG,RESP_TIME_MS,s2opt))
	< 0)  {
	if (rlen == S2_TIMEOUT)
	    goto E_TIMEOUT;			/* time out error */
	else
	    goto E_NETWORK;			/* other network error */
    }
    /* s2_close( sd, s2opt ); */
    
    if (debugon >= 2) plcdumpdata(resp,rlen,"messge received");
    
    if(ntohs(resp[0]) != 0xf100)  {		/* failure reponse */
	goto E_NETWORK2;			/* other network error */
    }

    {
	int i;
	for (i=0; i<size; i++)  {
	    data[i] = ntohs(resp[i+5]);
	}
    }

    return 0;

 E_SIZE:
    if(debugon)
	printf("plcread -- illegal data size '%d'\n", size);
    return -1;
    
 E_ADDR:
    if(debugon)
	printf("plcread -- illegal address specification '%s'\n", caddr);
    return -2;
    
 E_OPEN:
    if(debugon)
	printf("plcread -- open failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -3;
    
 E_TIMEOUT:
    if(debugon)
	printf("plcread -- timeout at send on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -4;

 E_NETWORK:
    if(debugon)
	printf("plcread -- plc network failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -5;

 E_NETWORK2:
    if(debugon)  {
	printf("plcread -- plc network internal failure on '%s'\n", host);
	printf(" -- terminal code is %x\n", ntohs(resp[0])&0xff);
    }
    s2_close( sd, s2opt );
    return -6;
}

/**********
 * plcwrite2 -- write values into a plc (WWR)
 */

int plcwrite2
(
 char *host,	/* host   - host name or ip address of yew plc */
 char *caddr,	/* caddr  - address to write ("Xxxxx") */
 int addr,	/* addr   - address to write (Dxxxx), when caddr is NULL */
 char *trans,	/* trans  - transport to use */
 int size,	/* size   - number of words to write (maximum of 256) */
 unsigned short *data	/* data   - data to write */
 )
/* externals: sd, resp, hostsd, s2opt, debugon */
{
    static char template[] = {			/* command to plc */
	0x12,	/* 0: write words sequentially		*/
	0x01,	/* 1: CPU No.				*/
	0x00,	/* 2: sentence consisting of (8 + 2n) bytes	*/
	0x00,	/* 3: 8+n*2  ( cmnd[4] -> cmnd[-] )	*/
	0x00, 0x04,		/* 4: kind of device = D (0x0004)	*/
	0, 0, 0x03, 0xe8,	/* 6: start at channel No.00001000	*/
	0x00, 0x09,		/* 10: n				*/
	0, 0,			/* 12: value 16bit			*/
	0, 0,			/* 14: value 16bit			*/
	0x0
    };
    static unsigned short cmnd[MAX_MSG/2];	/* command for plc */
    int slen = size*2+12;		/* command length */
    int rlen;				/* reply length */

    if ((size>MAX_MSG/4)||(size<1))  goto E_SIZE;

    if ( cmnd[0] != htons(0x1201) )  {		/* first time */
	cmnd[0] = htons(0x1201);		/* command & cpu number */
	cmnd[3] = htons(0x0000);		/* data address uppper word */
    }
    if ( sd == -1 || strncmp(host,hostsd,64) )  {
	s2_close(sd,S2_OERR);
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt )) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	strncpy(hostsd,host,64);		/* save hostname */
    }
    
    cmnd[1] = htons(8+size*2);			/* message size */
    cmnd[5] = htons(size);			/* data number */
    
    if (caddr != NULL)  {			/* address in "Xxxxx" format */
	char c;
	int d;
	if ( 2 != sscanf(caddr, "%c%d", &c, &d) )  
	    goto E_ADDR;			/* address format */
	cmnd[2] = htons(c-'@');			/* data class */
	cmnd[4] = htons(d);			/* data address */
    }  else  {					/* Dxxxx address */
	cmnd[2] = htons(0x0004);		/* data class */
	cmnd[4] = htons(addr);			/* data address */
    }
    
    {
	int i;
	for (i=0; i<size; i++)  {
	    cmnd[i+6] = htons(data[i]);
	}
    }

    if ( (rlen = s2_tsend(sd, (char *)cmnd, slen, RESP_TIME_MS, s2opt))
	!= slen){
	s2_close(sd,S2_OERR);			/* try again */
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt)) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	    != slen )  {
	    if (rlen == S2_TIMEOUT)  {		/* time out ? */
		goto E_TIMEOUT;			/* time out error */
	    }  else  {
		goto E_NETWORK;			/* other network error */
	    }
	}
    }
    if ( (rlen = s2_treceive(sd,(char *)resp,MAX_MSG,RESP_TIME_MS,s2opt))
	< 0)  {
	if (rlen == S2_TIMEOUT)
	    goto E_TIMEOUT;			/* time out error */
	else
	    goto E_NETWORK;			/* other network error */
    }
    /* s2_close(sd,s2opt); */
    
    if(ntohs(resp[0]) != 0x9200){		/* failure reponse */
	goto E_NETWORK2;			/* other network error */
    }
    return 0;

 E_SIZE:
    if(debugon)
	printf("plcwrite -- illegal data size '%d'\n", size);
    return -1;
    
 E_ADDR:
    if(debugon)
	printf("plcwrite -- illegal address specification '%s'\n", caddr);
    return -2;
    
 E_OPEN:
    if(debugon)
	printf("plcwrite -- open failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -3;
    
 E_TIMEOUT:
    if(debugon)
	printf("plcwrite -- timeout at send on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -4;

 E_NETWORK:
    if(debugon)
	printf("plcwrite -- plc network failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -5;

 E_NETWORK2:
    if(debugon)  {
	printf("plcwrite -- plc network internal failure on '%s'\n", host);
	printf(" -- terminal code is %x\n", ntohs(resp[0])&0xff);
    }
    s2_close( sd, s2opt );
    sd = -1;
    return -6;
}


/**********
 * plcread2 -- read values from a plc (WRD)
 */
int plcread2
(
 char *host,	/* host   - host name or ip address of yew plc */
 char *caddr,	/* caddr  - address to read ("Yxxxx") */
 int addr,	/* addr   - address to read (Dxxxx), when caddr is NULL */
 char *trans,	/* trans  - transport to use */
 int size,	/* size   - number of words to read (maximum of 247?) */
 unsigned short *data	/* data   - data to read */
 )
/* externals: sd, resp, hostsd, s2opt, debugon */
{
    static char template[] = {			/* command to plc */
	0x11,	/* 0: read words sequentially	*/
	0x01,	/* 1: CPU No.					*/
	0x00,	/* 2: sentence consisting of 18 bytes*/
	0x08,	/* 3: 8 ( cmnd[4] -> cmnd[-] )	*/
	0x00, 0x04,		/* 4: kind of device = D (0x0004)	*/
	0, 0, 0x03, 0xe8,	/* 6: start at channel, ex. No.00005000 */
	0x00, 0x00		/* 10: n */
    };
    static unsigned short cmnd[MAX_MSG/2]; /* command for plc */
    int slen = 12;				/* command length */
    int rlen;					/* reply length */

    if ((size>MAX_MSG/4)||(size<1))  goto E_SIZE;

    if ( cmnd[0] != htons(0x1101) )  {		/* first time */
	cmnd[0] = htons(0x1101);		/* command & cpu number */
	cmnd[1] = htons(0x0008);		/* message size */
	cmnd[3] = htons(0x0000);		/* data address upper word */
    }
    if ( sd == -1 || strncmp(host,hostsd,64) )  {
	s2_close(sd,S2_OERR);
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt )) < 0)  {
	    goto E_OPEN;
	}
	strncpy(hostsd,host,64);		/* save hostname */
    }

    cmnd[5] = htons(size);			/* data number */

    if (caddr != NULL)  {			/* address in "Xxxxx" format */
	char c;
	int d;
	if ( 2 != sscanf(caddr, "%c%d", &c, &d) )
	    goto E_ADDR;			/* address format */
	cmnd[2] = htons(c-'@');			/* data class */
	cmnd[4] = htons(d);			/* data address */
    }  else  {
	cmnd[2] = htons(0x0004);		/* data class */
	cmnd[4] = htons(addr);			/* data address */
    }

    if (debugon >= 2) plcdumpdata(cmnd,slen,"messge to be sent");
    
    if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	!= slen)  {
	s2_close(sd,S2_OERR);			/* try again */
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt )) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	    != slen)  {
	    if (rlen == S2_TIMEOUT)  {		/* time out ? */
		goto E_TIMEOUT;			/* time out error */
	    }  else  {
		goto E_NETWORK;			/* other network error */
	    }
	}
    }
    if ((rlen = s2_treceive(sd,(char *)resp,MAX_MSG,RESP_TIME_MS,s2opt))
	< 0)  {
	if (rlen == S2_TIMEOUT)
	    goto E_TIMEOUT;			/* time out error */
	else
	    goto E_NETWORK;			/* other network error */
    }
    /* s2_close( sd, s2opt ); */
    
    if (debugon >= 2) plcdumpdata(resp,rlen,"messge received");
    
    if(ntohs(resp[0]) != 0x9100)  {		/* failure reponse */
	goto E_NETWORK2;			/* other network error */
    }

    {
	int i;
	for (i=0; i<size; i++)  {
	    data[i] = ntohs(resp[i+2]);
	}
    }

    return 0;

 E_SIZE:
    if(debugon)
	printf("plcread -- illegal data size '%d'\n", size);
    return -1;
    
 E_ADDR:
    if(debugon)
	printf("plcread -- illegal address specification '%s'\n", caddr);
    return -2;
    
 E_OPEN:
    if(debugon)
	printf("plcread -- open failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -3;
    
 E_TIMEOUT:
    if(debugon)
	printf("plcread -- timeout at send on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -4;

 E_NETWORK:
    if(debugon)
	printf("plcread -- plc network failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -5;

 E_NETWORK2:
    if(debugon)  {
	printf("plcread -- plc network internal failure on '%s'\n", host);
	printf(" -- terminal code is %x\n", ntohs(resp[0])&0xff);
    }
    s2_close( sd, s2opt );
    return -6;
}

/**********
 * plcquery -- query to a plc (INF)
 */
int plcquery
(
 char *host,	/* host   - host name or ip address of yew plc */
 char *trans,	/* trans  - transport to use */
 int option,	/* option - query option, 1 - 5 */
 int unit,	/* unit   - 0 - 7 */
 unsigned short *data,	/* data   - data to read */
 int *size	/* size   - received data size */
 )
{
    static char template[] = {			/* command to plc */
	0x62,	/* 0: read general information		*/
	0x01,	/* 1: CPU No.				*/
	0x00,	/* 2: sentence consisting of 4 bytes	*/
	0x04,	/* 3: 4 ( cmnd[4] -> cmnd[-] )		*/
	0x00,	/* 4: option */
	0x00,	/* 5: option */
	0x00,	/* 6: unit */
	0x00,	/* 7: unit */
    };
    static unsigned short cmnd[MAX_MSG/2]; /* command for plc */
    int slen = 8;				/* command length */
    int rlen;					/* reply length */

    if ((option>5)||(option<1))  goto E_OPTION;

    if ( cmnd[0] != htons(0x6201) )  {		/* first time */
	cmnd[0] = htons(0x6201);		/* command & cpu number */
	cmnd[1] = htons(0x0004);		/* message size */
    }
    if ( sd == -1 || strncmp(host,hostsd,64) )  {
	s2_close(sd,S2_OERR);
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt )) < 0)  {
	    goto E_OPEN;
	}
	strncpy(hostsd,host,64);		/* save hostname */
    }

    cmnd[2] = htons(option);			/* option */
    cmnd[3] = htons(unit);			/* unit */

    if (debugon >= 2) plcdumpdata(cmnd,slen,"messge to be sent");
    
    if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	!= slen)  {
	s2_close(sd,S2_OERR);			/* try again */
	if((sd = s2_open(YEW_PL_PORT,"",trans,host,s2opt )) < 0)  {
	    goto E_OPEN;			/* open communication */
	}
	if ((rlen = s2_tsend(sd,(char *)cmnd,slen,RESP_TIME_MS,s2opt))
	    != slen)  {
	    if (rlen == S2_TIMEOUT)  {		/* time out ? */
		goto E_TIMEOUT;			/* time out error */
	    }  else  {
		goto E_NETWORK;			/* other network error */
	    }
	}
    }
    if ((rlen = s2_treceive(sd,(char *)resp,MAX_MSG,RESP_TIME_MS,s2opt))
	< 0)  {
	if (rlen == S2_TIMEOUT)
	    goto E_TIMEOUT;			/* time out error */
	else
	    goto E_NETWORK;			/* other network error */
    }
    /* s2_close( sd, s2opt ); */
    
    if (debugon >= 2) plcdumpdata(resp,rlen,"messge received");
    
    if(ntohs(resp[0]) != 0xe200)  {		/* failure reponse */
	goto E_NETWORK2;			/* other network error */
    }

    *size = ntohs(resp[1]);
    {
	int i;
	/* for (i=0; i<MAX_QUERY; i++)  { */
	for (i=0; i<*size/2; i++)  {
	    /* data[i] = ntohs(resp[i+2]); */
	    data[i] = resp[i+2];		/* byte data */
	}
    }

    return 0;

 E_OPTION:
    if(debugon)
	printf("plcread -- illegal option '%d'\n", option);
    return -1;
    
 E_OPEN:
    if(debugon)
	printf("plcread -- open failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -3;
    
 E_TIMEOUT:
    if(debugon)
	printf("plcread -- timeout at send on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -4;

 E_NETWORK:
    if(debugon)
	printf("plcread -- plc network failure on '%s'\n", host);
    s2_close( sd, s2opt );
    sd = -1;
    return -5;

 E_NETWORK2:
    if(debugon)  {
	printf("plcread -- plc network internal failure on '%s'\n", host);
	printf(" -- terminal code is %x\n", ntohs(resp[0])&0xff);
    }
    s2_close( sd, s2opt );
    return -6;
}

int plcdumpdata(unsigned short *resp, int len, char *mess)
{
    int i;
    
    printf("## %s ##\n", mess);
    for (i=0; i<len/2; i++) {
	printf("%4.4x,",resp[i]);
    }
    printf("\n");
    return 0;
}


#ifdef MAIN
char *nm;					/* program name */
int verbose = 0;				/* verbose flag, default off */
int loopcount = INT_MAX;			/* loop count, default inf. */
int interval = 1;				/* interval, default 1 sec */
int protocol = 0;				/* protocol, wrd/wwr:0 mcm:7 */
int numread = 10;				/* no. of data, default 10 */
int numwrite = 0;				/* no. of data to write */
int increment = 0;				/* increment flag */
int queryopt = 0;				/* query option */
int queryopt2 = 0;				/* query option 2 */
char * hostname = NULL;				/* hostname */
char * hostname2 = NULL;			/* 2nd hostname */
char * straddr = NULL;				/* address */
char * udptcp = "udp";				/* default transport is udp */
unsigned short databuf[MAX_MSG/2];

/*
 * argument -- process argument
 */
void argument(int argc, char **argv) {
    int c,data,dataflag=0;
    extern int optind, opterr;
    extern char *optarg;
    opterr = 0;					/* disable warning message */

    while((c = getopt (argc, argv, "dDl:L:i:I:h:H:g:G:a:A:p:P:r:R:w:W:c:C:mMq:Q:tT")) != EOF)
	switch ( c )  {
	case 'd':
	case 'D':
	    debugon++;
	    verbose++;  break;			/* verbose */
	case 'l':
	case 'L':
	    loopcount = atoi( optarg );  break;	/* loop count */
	case 'i':
	case 'I':
	    interval = atoi( optarg );  break;	/* interval */
	case 'h':
	case 'H':
	    hostname = optarg;  break;		/* plc hostname or ip */
	case 'g':
	case 'G':
	    hostname2 = optarg;  break;		/* plc hostname or ip */
	case 'a':
	case 'A':
	    straddr = optarg;  break;		/* address Xxxxx */
	case 'p':
	case 'P':
	    protocol = atoi( optarg );  break;	/* protocol to use */
	case 'r':
	case 'R':
	    numread = atoi( optarg );  break;	/* number of data to read */
	case 'w':
	case 'W':
	    numwrite = atoi( optarg );  break;	/* number of data to write */
	case 'c':
	case 'C':
	    dataflag++;
	    data = atoi( optarg );  break;	/* data to write */
	case 'm':
	case 'M':
	    increment++;  break;		/* increment each loop */
	case 'q':
	case 'Q':
	    queryopt = atoi( optarg );		/* query option */
	    {
		char *c;
		if (c=strpbrk(optarg,":;,/")) queryopt2=atoi(c+1);
		else queryopt2 = 0;
	    }  break;		/* specification may be "-q 3:0" or "-q3" */
	case 't':
	case 'T':
	    udptcp = "tcp";  break;		/* use tcp protocol */
	default:
	    fprintf( stderr, "%s: invalid option='%c'\n", nm, c );
	    goto usage;				/* print out usage */
	}

    switch (argc-optind)  {
    case 0:
	if (numwrite)  {
	    if (dataflag)  {
		int i;
		for (i=0; i<numwrite; i++) databuf[i] = data;
	    }  else  {
		fprintf( stderr, "%s: no data specified with -w\n", nm );
		goto usage;			/* print out usage */
	    }
	}
	if (!numread && !queryopt)  {		/* -r 0 and no -q */
	    fprintf( stderr, "%s: zero specified for -r\n", nm );
	    goto usage;			/* print out usage */
	}
	break;
    default:
	{
	    int i;
	    numwrite = argc-optind;
	    for (i=0; i<numwrite; i++)  {
		databuf[i] = strtol( argv[optind+i], NULL, 0 );
	    }
	}
	break;
    }
    if (!hostname)  {				/* no -h */
	fprintf( stderr, "%s: no plc host specified\n", nm );
	goto usage;				/* print out usage */
    }
    if (!straddr && !queryopt)  {			/* no -a and no -q */
	fprintf( stderr, "%s: no address specified\n", nm );
	goto usage;				/* print out usage */
    }
    return;

 usage:
    fprintf(stderr, "\n%s\n", what_plc4);
    fprintf(stderr, "read/write plc (test program)\n");
    fprintf(stderr, 
"Usage:   plc4 [<options>] [data1 [data2 [...]]]\n\
 options:\n\
  -d  debug information\n\
  -dd more verbose\n\
  -l {C}  loop count\n\
  -i {S}  loop interval\n\
  -m  increment each data at each loop\n\
  -h {host}   plc hostname\n\
  -g {host2}  optional 2nd plc hostname\n\
  -a {Xxxxx}  plc register address\n\
  -r {N}  number of data to read\n\
  -w {N}  number of data to write\n\
  -c {data}  data to write (using with -w)\n\
  -p {prot}  protocol, wwr/wrd:0 (default), mcm:7\n\
  -t  use tcp transport instead of udp\n\
  -q {opt}  query option\n\
   '-h host' and ('-a' or '-q') are mandatory.\n\
   data1... are data to write (0xddd for hex)\n");
    fprintf(stderr, 
" examples:\n\
  plc4 -h yewplc -a D100 -w 10 -c 3 -l 10\n\
  plc4 -h yewplc -a D200 -r 3 0xff 11 0xee\n\
  plc4 -h yewplc -a D300 -r 3 -m -l 100 -i 0 10 20 30\n\
  plc4 -h yewplc -g yewplc2 -a D300 -r 3 -l 1 -i 0\n");
    fprintf(stderr, 
" options for query\n\
  1        cpu status, and program status\n\
  2        system id., cpu type, and area size\n\
  3:{unit} module names, and input/output info.\n\
 examples:\n\
  plc4 -h yewplc -q 1 -i10\n\
  plc4 -h yewplc -q 2\n\
  plc4 -h yewplc -q 3:0 -i0 -l1\n\
  plc4 -h yewplc -q 3:1\n");
    exit( 1 );
}

#define MAX_SIZE 512

void update( char *host, char *addr, char *trans, int numread, 
	    int numwrite, unsigned short *data )
{
    unsigned short rdata[MAX_SIZE];
    int i;

    if ( protocol == 0 )  {
	if ( plcread2( host, addr, 0, trans, numread, rdata ) )  {
	    fprintf(stderr, " plcread2 error\n" );
	    exit( 1 );
	}
    }  else  {
	if ( plcread4( host, addr, 0, trans, numread, rdata ) )  {
	    fprintf(stderr, " plcread4 error\n" );
	    exit( 1 );
	}
    }
    fflush(stderr);
    fflush(stdout);
    for (i=0; i<numread; i++)  printf("%6d", rdata[i]);
    printf("\n");
    fflush(stdout);
#ifndef NOWRITE
    if (numwrite)  {
	if ( protocol == 0 )  {
	    if ( plcwrite2( host, addr, 0, trans, numwrite, data ) )  {
		fprintf(stderr, " plcwrite2 error\n" );
		exit( 1 );
	    }
	}  else  {
	    if ( plcwrite4( host, addr, 0, trans, numwrite, data ) )  {
		fprintf(stderr, " plcwrite4 error\n" );
		exit( 1 );
	    }
	}
    }
    
/*
    {
	int tmp = data[0];
	for (i=0; i<numdata-1; i++) data[i] = data[i+1];
	data[numdata-1] = tmp;
    }
*/
#endif /* NOWRITE */
    return;
}

void query( char *host, char *trans, int option, int option2 )
{
    unsigned short rdata[MAX_SIZE];
    int i;
    int size;

    if ( plcquery( host, trans, option, option2, rdata, &size ) )  {
	fprintf(stderr, " plcquery error\n" );
	exit( 1 );
    }
    fflush(stderr);
    fflush(stdout);
    switch (option)  {
    case 1:
	{
	    unsigned char *c=(unsigned char *)&(rdata[0]);
	    static char *cpu[] = {"Ready\n","Not-Ready\n"};
	    static char *prog[] = {"-","Running\n","Stopped\n","Debugging\n"};
	    fputs("CPU Status:     ",stdout);
	    switch (*c)  {
	    case 0:
	    case 1:
		fputs(cpu[*c],stdout);
		break;
	    default:
		fprintf(stdout,"unknown status %d\n",*c);
	    }
	    fputs("Program Status: ",stdout);
	    switch (*++c)  {
	    case 0xff:
		fputs("ROM Writer\n",stdout);
		break;
	    case 1:
	    case 2:
	    case 3:
		fputs(prog[*c],stdout);
		break;
	    default:
		fprintf(stdout,"unknown status %d\n",*c);
	    }
	}
	break;
    case 2:
	{
	    int i;
	    unsigned char *c=(unsigned char *)&(rdata[0]);
	    printf("System ID:     %-16.16s\n",c);
	    printf("Revision:      %-8.8s\n",c+16);
	    i=ntohs(rdata[12])&0x3;
	    if (i==1) printf("CPU Type:      Sequence\n");
	    else if (i==2) printf("CPU Type:      BASIC\n");
	    else printf("CPU Type:      unknown\n");
	    printf("Program Area:  %d k-step\n",ntohs(rdata[13]));
	}
	break;
    case 3:
	for (i=0; i<size/2; i+=3)  {
	    unsigned char *c=(unsigned char *)&(rdata[i]);
	    int j;
	    static char *io[] = {"  -  ", "  in-", " out-", " i/o-"};
		
	    for (j=0; j<4; j++, c++)  fputc(isprint(*c)?*c:'-',stdout);
	    if (*c == 0)  {
		fputs("        \n",stdout);
		c+=2;
	    } else {
		fputs(io[*c++&3],stdout);
		fprintf(stdout,"%-3u\n",(*c++)&0xff);
	    }
	}
	break;
    case 4:
	
    default:
	fprintf( stderr, "%s: illegal option specified for -q\n", nm );
	break;
    }
    fflush(stdout);
    return;
}

int main(int argc, char **argv) {
    nm = argv[0];				/* save program name */
    hostname2 = NULL;

    argument(argc, argv);			/* process arguments */

    /* if (!interval)  interval++; */		/* allow zero interval */
    if (!numread)  numread = numwrite;
    if (verbose)  {
	fprintf(stderr, "loop count = %d\n", loopcount );
	fprintf(stderr, "sleep interval = %d\n", interval );
	if (protocol == 0) 
	    fprintf(stderr, "protocol to be used is WRD/WWR\n");
	if (protocol == 7) 
	    fprintf(stderr, "protocol to be used is MCM\n");
	if (!strcmp(udptcp,"udp"))
	    fprintf(stderr, "transport to be used is UDP\n");
	if (!strcmp(udptcp,"tcp"))
	    fprintf(stderr, "transport to be used is TCP\n");
	fprintf(stderr, "number of read data = %d\n", numread );
	fprintf(stderr, "number of write data = %d\n", numwrite );
	if (numwrite)  {
	    int i;
	    fprintf(stderr, "data to be written:\n");
	    for (i=0; i<numwrite; i++)  {
		fprintf(stderr, "%6d", databuf[i]);
	    }
	    fprintf(stderr, "\n");
	}
	s2opt = S2_OALL;
    }

    {						/* repeat */
	int i;
	for (i=0; i<loopcount; i++)  {		/* process */
	    if (queryopt)  {
		query( hostname, udptcp, queryopt, queryopt2 );
	    }  else  {
		update( hostname, straddr, udptcp, numread, numwrite, databuf );
		if (hostname2) {
		    update( hostname2, straddr, udptcp, numread, numwrite, databuf );
		}		
		if (numwrite && increment)  {
		    int i;
		    for (i=0; i<numwrite; i++) databuf[i]++;
		}
	    }
	    if (interval) sleep( interval );
	}
    }
    plcclose();
    
    exit( 0 );
}

#endif /* MAIN */
